import re
import sys
import os
sys.path.append(os.path.split(os.path.abspath(__file__))[0][:-3])
import cobra
from ..classes.model import model
from .Excel import ReadExcel, WriteExcel
from .ScrumPy import ReadScrumPyModel, WriteScrumPyModel
from past.builtins import basestring

def ReadModel(model_file=None, model_format=None, excel_parse="cobra_string",
          variable_name=None, Print=False, compartment_dic={}, bounds=1000.0, **kwargs):
    """ model_format = "sbml" | "sbml_legacy" | "excel" | "matlab" | "json" | "scrumpy | "yaml"
        excel_parse = "cobra_string" | "cobra_position" """
    if not model_file:
        pass
    elif model_format == "sbml" or model_format == "xml" or (
        model_format == None and model_file.endswith(".sbml")) or (
        model_format == None and model_file.endswith(".xml")):
        model_file = cobra.io.read_sbml_model(model_file, **kwargs)
    elif model_format == "sbml_legacy":
        model_file = cobra.io.read_legacy_sbml(model_file, **kwargs)
#        model_file = cobra.io.read_sbml_model(model_file,
#            old_sbml=old_sbml, legacy_metabolite=legacy_metabolite,
#            print_time=False, use_hyphens=use_hyphens)
    elif model_format == "matlab" or (model_format == None and
        model_file.endswith(".mat")):
        model_file = cobra.io.load_matlab_model(model_file,
                                                variable_name=variable_name, **kwargs)
    elif model_format == "json" or (model_format == None and
                                            model_file.endswith(".json")):
        model_file = cobra.io.load_json_model(model_file, **kwargs)
    elif model_format == "excel" or model_format == "xls" or \
        model_format == "cobra" or (model_format == None and
        model_file.endswith(".xls")) or (model_format == None
                                    and model_file.endswith(".xlsx")):
        model_file = ReadExcel(model_file,
                                       parse=excel_parse, Print=Print, **kwargs)
    elif model_format == "scrumpy" or model_format == "spy" or (
        model_format == None and model_file.endswith(".spy")):
        model_file = ReadScrumPyModel(model_file,
                    compartment_dic=compartment_dic, Print=Print, **kwargs)
    elif model_format == "yaml" or model_file.endswith(".yaml") or model_file.endswith(".yml"):
        model_file = cobra.io.load_yaml_model(model_file, **kwargs)
    m = model(model_file)
    if isinstance(model_file, basestring):
        m.description = m.id = model_file.rsplit(".",1)[0]
    m.SetBounds(bounds=bounds)
    return m

def WriteModel(model, filename, model_format=None, excel_format="cobra", ExtReacs=[], **kwargs):
    """ model_format = "sbml" | "sbml_legacy" | "excel" | "matlab" | "json" | "cobra" | "cobra_old" | "scrumpy" | "yaml" """
    if model_format == "sbml" or model_format == "xml"or (
        model_format == None and filename.endswith(".sbml")) or (
        model_format == None and filename.endswith(".xml")):
        cobra.io.write_sbml_model(model, filename, **kwargs)
    elif model_format == "sbml_legacy":
        description = model.description
        model.description = str(description)
        m_id = model.id
        model.id = str(m_id)
        compartments = model.compartments
        if not model.compartments:
            model.compartments = {'':''}
        original_met_id = {}
        original_met_compartment = {}
        for met in model.metabolites:    # changes metabolites in the model
            original_met_id[met] = met.id
            original_met_compartment[met] = met.compartment
            met.id = re.sub('[-/().,\[\]+]','_',met.id)
            if not met.compartment:
                met.compartment = ''
        original_reac_id = {}
        for reac in model.reactions:
            original_reac_id[reac] = reac.id
            reac.id = re.sub('[-/().,\[\]+]','_',reac.id)
        # cobra.io.write_sbml_model(model, filename=filename,
        #     sbml_level=sbml_level, sbml_version=sbml_version,
        #     print_time=False, use_fbc_package=fbc)
        cobra.io.write_legacy_sbml(model, filename=filename, **kwargs)
        model.description = description
        model.id = m_id
        model.compartments = compartments
        for met in model.metabolites:    # revert changes to metabolites
            met.id = original_met_id[met]
            met.compartment = original_met_compartment[met]
        for reac in model.reactions:
            reac.id = original_reac_id[reac]
    elif model_format == "matlab" or (
                model_format == None and filename.endswith(".mat")):
        cobra.io.save_matlab_model(model, filename, **kwargs)
    elif model_format == "json" or (
                model_format == None and filename.endswith(".json")):
        cobra.io.save_json_model(model, filename, **kwargs)
    elif model_format == "excel" or model_format == "xls" or (
        model_format == None and filename.endswith(".xls")) or (
        model_format == None and filename.endswith(".xlsx")):
        WriteExcel(model, filename, excel_format=excel_format)
    elif model_format == "cobra":
        WriteExcel(model, filename, excel_format="cobra")
    elif model_format == "cobra_old":
        WriteExcel(model, filename, excel_format="cobra_old")
    elif model_format == "scrumpy" or model_format == "spy" or (
                model_format == None and filename.endswith(".spy")):
        WriteScrumPyModel(model, filename, ExtReacs=ExtReacs)
    elif model_format == "yaml" or filename.endswith(".yaml") or filename.endswith(".yml"):
        cobra.io.save_yaml_model(model, filename, **kwargs)
    else: 
        print('Please specify model_format')



from .matrix import matrix
import numpy

class fca(matrix):

    def __init__(self,*args,**kwargs):
        super(matrix,self).__init__(*args,**kwargs)

    def Copy(self):
        return fca(self.copy())

#    def FCAGetReac(self,reac):
#        rv = fca(columns=self.columns)
#        row = []
#        reacs = self.columns.tolist()
#        idx = self.rnames.index(reac)
#        for b in range(len(reacs)):
#            row.append(self[idx][b])
#        rv.NewRow(r=row,name=reac)
#        return fca(self[reac])


    def ReacCoupling(self,reac,coupling="all"):
        reacds = fca(self[reac])
        fc = reacds.FullyCoupled()
        pc = reacds.PartiallyCoupled()
        dc = reacds.DirectionallyCoupled()
        rv = {"all":(fc,pc,dc),"FC":fc,"PC":pc,"DC":dc}
        return rv[coupling]


    def DirectionallyCoupled(self):
        implies = {}    # reac implies reacs, r1 -> r2, equivalent knockout
        impliedby = {}    # reac implied by reacs, r1 <- r2, affected
        txds = self.Copy()
        #txfs = fca(txds.transpose())
        for r in range(len(self.index)):
            rreac = self.index[r]
            impdby = []
            for c in range(len(self.columns)):
                creac = self.columns[c]
                if self.iloc[r,c] == 7:
                    impdby.append((creac,"+"))
                elif self.iloc[r,c] == 8:
                    impdby.append((creac,"-"))
            if len(impdby) > 0:
                impliedby[rreac] = impdby
        for r in range(len(txds.index)):
            rreac = txds.index[r]
            imps = []
            for c in range(len(txds.columns)):
                creac = txds.columns[c]
                if txds.iloc[r,c] == 7:
                    imps.append((creac,"+"))
                elif txds.iloc[r,c] == 8:
                    imps.append((creac,"-"))
            if len(imps) > 0:
                implies[rreac] = imps
        rv = {"Implies":implies,"Implied by":impliedby}
        return rv


    def PartiallyCoupled(self):
        pc = {}
        n = 1
        for r in range(len(self.index)):
            rreac = self.index[r]
            for c in range(len(self.columns)):
                creac = self.columns[c]
                if self.iloc[r,c] == 11 or self.iloc[r,c] == 12:
                    if (rreac not in pc.keys()) and (creac not in pc.keys()):
                        pc[rreac] = "PC"+str(n),"+"
                        if self.iloc[r,c] == 11:
                            pc[creac] = "PC"+str(n),"+"
                        elif self.iloc[r,c] == 12:
                            pc[creac] = "PC"+str(n),"-"
                        n += 1
                    elif (rreac in pc.keys()) and (creac not in pc.keys()):
                        num,sign = pc[rreac]
                        if self.iloc[r,c] == 12:
                            sign = self.SwapSign(sign)
                        pc[creac] = num,sign
                    elif (rreac not in pc.keys()) and (creac in pc.keys()):
                        num,sign = pc[creac]
                        if self.iloc[r,c] == 12:
                            sign = self.SwapSign(sign)
                        pc[rreac] = num,sign
        rv = {}
        for reac in pc:
            num,sign = pc[reac]
            if rv.has_key(num):
                rv[num].append((reac,sign))
            else:
                rv[num] = [(reac,sign)]
        return rv


    def FullyCoupled(self):
        fc = {}
        n = 1
        for r in range(len(self.index)):
            rreac = self.index[r]
            for c in range(len(self.columns)):
                creac = self.columns[c]
                if self.iloc[r,c] == 13 or self.iloc[r,c] == 14:
                    if (rreac not in fc.keys()) and (creac not in fc.keys()):
                        fc[rreac] = "FC"+str(n),"+"
                        if self.iloc[r,c] == 13:
                            fc[creac] = "FC"+str(n),"+"
                        elif self.iloc[r,c] == 14:
                            fc[creac] = "FC"+str(n),"-"
                        n += 1
                    elif (rreac in fc.keys()) and (creac not in fc.keys()):
                        num,sign = fc[rreac]
                        if self.iloc[r,c] == 14:
                            sign = self.SwapSign(sign)
                        fc[creac] = num,sign
                    elif (rreac not in fc.keys()) and (creac in fc.keys()):
                        num,sign = fc[creac]
                        if self.iloc[r,c] == 14:
                            sign = self.SwapSign(sign)
                        fc[rreac] = num,sign
        rv = {}
        for reac in fc:
            num,sign = fc[reac]
            if rv.has_key(num):
                rv[num].append((reac,sign))
            else:
                rv[num] = [(reac,sign)]
        return rv

    def SwapSign(self,sign):
        if sign == "+":
            rv = "-"
        elif sign == "-":
            rv = "+"
        else:
            print("swap sign error")
        return rv

    def RangeType(self,lb,ub,tol,bounds):
        def Eq(a,b,rtol=tol,atol=tol):
            return numpy.allclose(a,b,rtol,atol)
        if lb != lb or ub != ub:    # check for nan
            print("nan problem")
            return float("NaN")
        else:
            if ub >= lb:
                lo = lb
                hi = ub
            elif lb > ub:
                hi = lb
                lo = ub
            else:
                print("range type lb ub error")
    #        if lo == float("-inf") and hi == float("inf"):
            if Eq(lo,-bounds) and Eq(hi,bounds):
                rv = 1    # uncoupled
    #        elif lo > float("-inf") and hi == float("inf"): # lower bound only
            elif lo > -bounds and Eq(hi,bounds): # lower bound only
                if abs(lo) < tol:
                    rv = 2
                elif lo < -tol:
                    rv = 4
                elif lo > tol:
                    rv = 9    # directionally coupled
    #        elif lo == float("-inf") and hi < float("inf"): # upper bound only
            elif Eq(lo,-bounds) and hi < bounds: # upper bound only
                if abs(hi) < tol:
                    rv = 3
                elif hi > tol:
                    rv = 5
                elif hi < -tol:
                    rv = 10   # directionally coupled
    #        elif lo > float("-inf") and hi < float("inf"):  # both bounded
            elif lo > -bounds and hi < bounds:  # both bounded
                if lo > -tol and hi > -tol: # same direction
                    if lo < tol and hi < tol:
                        rv = 15
                    elif lo < tol and hi >= tol:
                        rv = 7    # directionally coupled
                    elif lo >= tol and hi >= tol:
                        if abs(lo-hi) > tol:
                            rv = 11   # partially coupled
                        else:
                            rv = 13   # fully coupled
                elif lo < tol and hi < tol: # opposite direction
                    if lo > -tol and hi > -tol:
                        rv = 15
                    elif lo <= -tol and hi > -tol:
                        rv = 8    # directionally coupled
                    elif lo <= -tol and hi <= -tol:
                        if abs(lo-hi) > tol:
                            rv = 12   # partially coupled
                        else:
                            rv = 14   # fully coupled
                else:
                    rv = 6    # ???...
            else:
                print("range type error")
            return rv




from ..classes import matrix

def ConstraintScan(model, cd, lo, hi, n_p, MinFlux=True, IncZeroes=True,cobra=True):
    """ scan one reaction flux
        pre: cd = sum of reaction fluxes dictionary """
    state = model.GetState()
    rv = matrix.matrix(columns=["Constraint","ObjVal"])
    lo = float(lo)
    hi = float(hi)
    inc = (hi - lo)/(n_p-1)
    cur = lo
    for n in range(n_p):
        model.SetSumReacsConstraint(cd, cur, name='ConstraintScan')
        if MinFlux:
            model.MinFluxSolve(cobra=cobra,DisplayMsg=False)
        else:
            model.Solve(False)
        sol = model.GetSol(IncZeroes=IncZeroes)

        if not model.Optimal():
            obval = float("NaN") # indicate fail
        else:
            obval =  model.GetObjVal()
        sol["ObjVal"] = obval
        sol["Constraint"] = cur
        rv = rv.UpdateFromDic(sol)
        cur += inc
        model.DelSumReacsConstraint('ConstraintScan')
    model.SetState(state)
    return rv

def RatioScan(model, reac1, reac2, n_p, lo=0, hi=1, flux_val=None,
              IncZeroes=True, rev=False):
    """ scan the ratio of two reaction fluxes
        pre: flux_val = a fixed flux for the sum of the two reactions """
    state = model.GetState()
    rv = matrix.matrix(columns=["Ratio","ObjVal"])
    lo = float(lo)
    hi = float(hi)
    inc = (hi - lo)/(n_p-1)
    cur = lo
    for n in range(n_p):
        ratiodic = {reac1:cur, reac2:1-cur}
        rationame = model.SetReacsFixedRatio(ratiodic, GetMetName=True)
        if flux_val != None:
            model.SetFixedFlux({rationame:flux_val})
        model.Solve(False)
        sol = model.GetSol(IncZeroes=IncZeroes)
        if not model.Optimal():
            obval = float("NaN") # indicate fail
        else:
            obval =  model.GetObjVal()
        sol["ObjVal"] = obval
        sol["Ratio"] = cur
        rv = rv.UpdateFromDic(sol)
        cur += inc
        model.DelReacsFixedRatio(fixedratio=rationame)
    model.SetState(state)
    return rv

def Constraint2DScan(model, cd1, lo1, hi1, cd2, lo2, hi2, n_p, IncZeroes=True):
    """ scan two reaction fluxes simultaneously """
    state = model.GetState()
    rv = matrix.matrix(columns=["Constraint1","Constraint2","ObjVal"])
    lo1 = float(lo1)
    hi1 = float(hi1)
    inc1 = (hi1 - lo1)/(n_p-1)
    cur1 = lo1
    lo2 = float(lo2)
    hi2 = float(hi2)
    inc2 = (hi2 - lo2)/(n_p-1)
    cur2 = lo2
    for n1 in range(n_p):
        model.SetSumReacsConstraint(cd1,cur1,'temp1')
        for n2 in range(n_p):
            model.SetSumReacsConstraint(cd2,cur2,'temp2')
            model.Solve(False)
            sol = model.GetSol(IncZeroes=IncZeroes)
            if not model.Optimal():
                obval = float("NaN") # indicate fail
            else:
                obval =  model.GetObjVal()
            sol["ObjVal"] = obval
            sol["Constraint1"] = cur1
            sol["Constraint2"] = cur2
            rv = rv.UpdateFromDic(sol)
            cur2 += inc2
            model.DelSumReacsConstraint('temp2')
        cur2 = lo2
        cur1 += inc1
        model.DelSumReacsConstraint('temp1')
    model.SetState(state)
    return rv

def ConstraintRandomMinFluxScan(model, cd, lo, hi, n_p, it, IncZeroes=True,
                                reacs=None, exc=[], processes=None):
    """ same as ConstraintScan except using RWFM rather than FBA
        pre: cd = sum of reaction fluxes dictionary """
    state = model.GetState()
    rv = matrix.matrix(columns=["Constraint"])
    lo = float(lo)
    hi = float(hi)
    inc = (hi - lo)/(n_p-1)
    cur = lo
    for n in range(n_p):
        model.SetSumReacsConstraint(cd,cur,name='ConstraintRandomMinFluxScan')
        mtx = model.RandomMinFlux(it=it, reacs=reacs, exc=exc,
                                    processes=processes)
        sol = mtx.AverageFlux()
        sol["Constraint"] = cur
        rv = rv.UpdateFromDic(sol)
        cur += inc
        model.DelSumReacsConstraint('ConstraintRandomMinFluxScan')
    model.SetState(state)
    return rv

def RatioRandomMinFluxScan(model, reac1, reac2, n_p, it, lo=0, hi=1, flux_val=None,
        IncZeroes=True, reacs=None, exc=[], rev=False,processes=None):
    """ same as RatioScan except using RWFM rather than FBA """
    state = model.GetState()
    rv = matrix.matrix(columns=["Ratio"])
    lo = float(lo)
    hi = float(hi)
    inc = (hi - lo)/(n_p-1)
    cur = lo
    for n in range(n_p):
        ratiodic = {reac1:cur,reac2:1-cur}
        rationame = model.SetReacsFixedRatio(ratiodic, GetMetName=True)
        if flux_val != None:
            model.SetFixedFlux({rationame:flux_val})
        mtx = model.RandomMinFlux(it=it, reacs=reacs, exc=exc,
                                processes=processes)
        sol = mtx.AverageFlux()
        sol["Ratio"] = cur
        rv = rv.UpdateFromDic(sol)
        cur += inc
        model.DelReacsFixedRatio(fixedratio=rationame)
    model.SetState(state)
    return rv

def WeightingScan(model,objdic,lo,hi,n_p):
    """ scan by changing the objective coefficients for a subset of
        reactions in the objective function """
    state = model.GetState()
    rv = matrix.matrix(columns=["Weighting"])
    lo = float(lo)
    hi = float(hi)
    inc = (hi - lo)/(n_p-1)
    cur = lo
    for n in range(n_p):
        newobjdic = {}
        for r in objdic.keys():
            newobjdic[r] = objdic[r]*cur
        model.SetObjective(newobjdic)
        model.Solve(False)
        sol = model.GetSol()
        sol["Weighting"] = cur
        rv = rv.UpdateFromDic(sol)
        cur += inc
    model.SetState(state)
    return rv


""" Not functional! To be modified """
def MatchScan(model,cd,clo,chi,md,vd,vlo,vhi,n_p,samedirec=True,count=50,display=False,tol=1e-6,IncZeroes=True):
    state = model.GetState()
    rv = matrix.matrix(columns=["Constraint","VaryFlux","ObjVal"])
    clo = float(clo)
    chi = float(chi)
    inc = (chi - clo)/(n_p-1)
    cur = clo

    for n in range(n_p):
        model.SetSumReacsConstraint(cd,cur,name="cd")
        varyflux = model.MatchFlux(md=md,vd=vd,lo=vlo,hi=vhi,samedirec=samedirec,count=count,display=display,tol=tol)
        if varyflux == None:
            model.DelSumReacsConstraint("cd")
            print("varyflux = None")
            break
        model.SetSumReacsConstraint(vd,varyflux,name="vd")
        model.Solve(False)
        sol = model.GetSol(IncZeroes=IncZeroes)
        if not model.Optimal():
            objval = float("NaN") # indicate fail
        else:
            objval = model.GetObjVal()
        sol["VaryFlux"] = varyflux*vd.values()[0]
        sol["ObjVal"] = objval
        sol["Constraint"] = cur
        rv = rv.UpdateFromDic(sol)
        cur += inc
        model.DelSumReacsConstraint("cd")
        model.DelSumReacsConstraint("vd")
    model.SetState(state)
    return rv


import numpy as np
import matplotlib.pyplot as plt
import xlsxwriter
from scipy.stats.stats import pearsonr
import pylab,math,os,itertools
from operator import itemgetter
import seaborn as sns

marker = itertools.cycle(('<', '>', 'D', 'H', '_', '^', 'd', 'h', 'o', 'p', 's', 'v', 'x', '|','+', '*', ',', '.', '1', '3', '2', '4', '8'))

def Plot_Res(res,ReacLst, XLSPath = None, X=None,Y=None, ShowLegend=False, ShowFig = False):
    
    book = xlsxwriter.Workbook(XLSPath)
    
    crs = book.add_worksheet('Reactions')

    

    if ShowFig and X==None:
        raise ValueError("Need X-axis to plot")
        
    flg = 0    
    if X:
        crs.write(0,0,X)
        for val in res.to_dict()[X].keys():
            crs.write(val+1,0,res.to_dict()[X][val])
        flg = 1
        
    if X in ReacLst:
        ReacLst.remove(X)
    
    for reac in ReacLst:
        
        if X:
            
            plt.plot(res.to_dict()[X].values(),ReacToFlux(res,reac),label=reac,marker=marker.next())
        
        crs.write(0,ReacLst.index(reac)+flg,reac)
        
        for val in res.to_dict()[reac].keys():
            
            crs.write(val+1 ,ReacLst.index(reac)+flg,res.to_dict()[reac][val])
        
    
    if X:
        plt.xlabel(X)
    if Y:
        plt.ylabel(Y)
    if ShowLegend:
        plt.legend()

    try:
        
        book.close()
    except:
        print("Need excel path to save data")
    
    if ShowFig:  
      plt.show()

    

    

def Constants(res,tot=1e-8, XLSPath = None, X=None,Y=None, ShowLegend=False, ShowFig = False):

    """res = DataSet = pandas DataFrame object. Plot, save and returns the reactions those remain constant (no variation) throughout.
        ShowFig requires 'X - axis'
        Modify tot for constant definition"""
    #example FluxAnalysis.Constants(res,XLSPath='/home/rahul/Desktop/Cons.xls',ShowFig=True,X='R1')
    
    print("This may take some time.....")
    
    ReacLst=[]
    for reac in res.to_dict().keys():
         
        flux = ReacToFlux(res,reac)
        if all(abs(flux[0]-i)<=tot for i in flux):
            ReacLst.append(reac)
            
    Plot_Res(res,ReacLst, XLSPath = XLSPath, X=X,Y=Y, ShowLegend=ShowLegend, ShowFig = ShowFig)

    
    return ReacLst



def ONRange(res,minx=0,maxx=10, XLSPath = None, X=None,Y=None, ShowLegend=False, ShowFig = False):
    """res = DataSet = pandas DataFrame object. Plot, save and returns reactions those are remain ON between minx to maxx times (occurrence)
    ShowFig requires 'X - axis'"""

    #example FluxAnalysis.ONRange(DataSet, minx=5,maxx=10,XLSPath='/home/rahul/Desktop/OccurON.xls') --> Reactions remain ON (>0) between 5 to 10 times (occurrence)
    
    ReacLst=[]
    for reac in res.to_dict().keys():
        
        flux = ReacToFlux(res,reac)
        if len(flux)-flux.count(0)>=minx and len(flux)-flux.count(0)<=maxx:
            ReacLst.append(reac)
            
    Plot_Res(res,ReacLst, XLSPath = XLSPath, X=X,Y=Y, ShowLegend=ShowLegend, ShowFig = ShowFig)
            
    return ReacLst






def PlotCorHisto(res,XReactionName, XLSPath=None, CorrWithCompart = None):
    """res = DataSet = pandas DataFrame object. Pearson's Correlation coefficient (r) with a particular reaction (XReactionName) to all other in the dataset.
    r (with Nan) and p-value will be saved in excel"""
    

    dic = {}

    book = xlsxwriter.Workbook(XLSPath,{'nan_inf_to_errors': True})
    crs = book.add_worksheet('pearson-r')
    crs.write(0,0,'Reaction')
    crs.write(0,1,'r')
    crs.write(0,2,'p-value')

    print("This may take some time.....")
    
    
    

    
    x = ReacToFlux(res,XReactionName)

    row = 1
    
    for reac in res.to_dict().keys():

        if CorrWithCompart and CorrWithCompart in reac:
        
        
            y = ReacToFlux(res,reac)
            
            r = pearsonr(x, y)

            if not math.isnan(r[0]):
                dic[reac]=r[0]
            
            crs.write(row,0,reac)
            crs.write(row,1,r[0])
            crs.write(row,2,r[1])
            row += 1
            
        if CorrWithCompart == None:
            
            y = ReacToFlux(res,reac)
            
            r = pearsonr(x, y)

            if not math.isnan(r[0]):
                dic[reac]=r[0]
            
            crs.write(row,0,reac)
            crs.write(row,1,r[0])
            crs.write(row,2,r[1])
            row += 1
            

        
    n, bins, patches = pylab.hist(dic.values()) # for bins
    plt.hist(dic.values(), bins=bins)
    plt.xlabel('Correlation coefficient (r)')
    plt.ylabel('Frequency (No. of Reaction)')
    plt.legend()
    plt.show()

        
        

def BuildClassReacMatrix():


    typeIDlst=[]
    with open (os.path.dirname(__file__) + '/../Data/classes.dat','r') as AraClassFile:
        for bl, gr in itertools.groupby(AraClassFile, lambda line: line.startswith('//\n')):
            lgr=list(gr)
            for line in lgr:
                s = line.split()
                if (s[0] == 'UNIQUE-ID' or s[0]=='TYPES') and s[2] not in typeIDlst:
                    typeIDlst.append(s[2])
    with open (os.path.dirname(__file__) + '/../Data/pathways.dat','r') as PathFile:
        for bl, gr in itertools.groupby(PathFile, lambda line: line.startswith('//\n')):
            lgr=list(gr)
            for line in lgr:
                s = line.split()
                if (s[0] == 'UNIQUE-ID' or s[0]=='TYPES') and s[2] not in typeIDlst:
                    typeIDlst.append(s[2])
    with open (os.path.dirname(__file__) + '/../Data/pathways.dat','r') as PathFile:
        ReacLst=[]
        for bl, gr in itertools.groupby(PathFile, lambda line: line.startswith('//\n')):
            lgr=list(gr)
            for line in lgr:
                s = line.split()
                if line.startswith('REACTION-LIST') and s[2] not in ReacLst:
                    ReacLst.append(s[2])

                    
    
    with open (os.path.dirname(__file__) + '/../Data/ClassReacMatrix.txt','w') as MatFile:
        MatFile.write('\t')
        for cls in typeIDlst:
            MatFile.write(cls+'\t')
        MatFile.write('\n')
        for reac in ReacLst:
            MatFile.write(reac+'\t')
            pl=[]
            with open(os.path.dirname(__file__) +'/../Data/pathways.dat','r') as PathFile:
                for bl, gr in itertools.groupby(PathFile, lambda line:line.startswith('//\n')):
                    lgr=list(gr)
                    if 'REACTION-LIST - '+reac+'\n' in lgr:
                        for line in lgr:
                            if line.startswith('TYPES - ') or line.startswith('UNIQUE-ID - '):
                                pl.append(line.split()[2])
            for pid in typeIDlst:
                if pid in pl:
                    MatFile.write('1\t')
                else:
                    MatFile.write('0\t')
            MatFile.write('\n')


def __PathReacMap(PathList):
    
    rv = []
    try:
        with open (os.path.dirname(__file__) + '/../Data/ClassReacMatrix.txt','r') as MatFile:
            pass
    except:
        print('\n\nClassReacMatrix.txt file not found\nScobra will automatically create this file\n')
        print('Once created, Scobra will use this file next time onwards\nFile creation requires classes.dat and pathways.dat in /Data directory of scobra\n')
        BuildClassReacMatrix()
        print('ClassReacMatrix.txt is successfully created - it holds Pathway and Reaction map\n')
    with open (os.path.dirname(__file__) + '/../Data/ClassReacMatrix.txt','r') as MatFile:
        for no, line in enumerate(MatFile):
            if no == 0:
                pcol = line.split()
            if no > 0:
                s=line.split()
                for path in PathList:
                    if s[pcol.index(path)+1]=='1' and s[0] not in rv:
                        
                        rv.append(s[0])

            
            
            
    return rv


def GetPathFlux(res,PathSuffixDic={},XLSPath = None, X=None,Y=None, ShowLegend=False, ShowFig = False):
    """res = DataSet = pandas DataFrame object. PathSuffixDic holds the pathway and suffix information, pathway is used to fetch pathway specific reactions, suffix is used to get localization.
    Any pathway 'ID' or 'Type' name can be used. Any parent or chlid class (classes.dat and pathways.dat file; http://pmn.plantcyc.org/ARA/class-tree?object=Pathways) or ID can be entered.
    ShowFig requires 'X - axis'
    same keys need to be affix __KEY__number"""

    #Combination provided below - can be used for central carbon metabolism , need edit??
    
    #PathSuffixDic={'CALVIN-PWY':'_p_Leaf_Day','GLYCOLYSIS-VARIANTS':'_c_Leaf_Day','Pentose-Phosphate-Cycle':'_p_Leaf_Day'} 

    
    #example rv=FluxAnalysis.GetPathFlux(DataSet,PathSuffixDic={'CALVIN-PWY':'_p_Leaf_Day','GLYCOLYSIS-VARIANTS':'_c_Leaf_Day','Pentose-Phosphate-Cycle':'_p_Leaf_Day'},XLSPath='/home/rahul/Desktop/Central_C.xls')
    #example rv=FluxAnalysis.GetPathFlux(DataSet,PathSuffixDic={'Energy-Metabolism':'_c_Leaf_Day'},XLSPath='/home/rahul/Desktop/Energy_at_Cytosol.xls',ShowFig=True,X='time')
    #example rv=FluxAnalysis.GetPathFlux(DataSet,PathSuffixDic={'Energy-Metabolism':'_p_Leaf_Day'},XLSPath='/home/rahul/Desktop/Energy_at_plastid.xls')
    #example rv=FluxAnalysis.GetPathFlux(res,PathSuffixDic={'GLYOXYLATE-BYPASS':'_c_Leaf_Day'},XLSPath='/home/rahul/Desktop/Misc.xls')

    ReacLst=[]
    rv={}
    for path in PathSuffixDic.keys():
        Rlist=__PathReacMap([path.split('_KEY_')[0]])
        UpdatedList=[i+PathSuffixDic[path] for i in Rlist]
        
        for reac in res.to_dict().keys():
            if reac in UpdatedList:
                flux = ReacToFlux(res,reac)
                
                    
                rv[reac]=flux
                if reac not in ReacLst:
                    ReacLst.append(reac)
    if XLSPath:           
        Plot_Res(res,ReacLst, XLSPath = XLSPath, X=X,Y=Y, ShowLegend=ShowLegend, ShowFig = ShowFig)

    return rv

        
def ReacToFlux(res,reac):
    
    ScanRange = res.to_dict()[reac]
    
    return  [val for ky, val in sorted(ScanRange.items())]
    



def PathHeatMap(res, PathSuffixDic={}, MeanPath=True,ReacAbssoluteVal=False,linewidths=0.05,Order=None):

    """ Plots heatmap of fluxes for individual reactions in a pathway or mean of absolute flux value for pathways
    ReacAbssoluteVal = True/False can only work when MeanPath = False"""
    
    #example FluxAnalysis.PathHeatMap(res,PathSuffixDic={'CALVIN-PWY':'_p_Leaf_Day','GLYCOLYSIS-VARIANTS':'_c_Leaf_Day','Pentose-Phosphate-Cycle':'_p_Leaf_Day'})  --> Central Carbon
    
    #example FluxAnalysis.PathHeatMap(res,PathSuffixDic={'CALVIN-PWY_KEY_1':'_p_Leaf_Day','CALVIN-PWY_KEY_2':'_p_Root_Night'},MeanPath=False,ReacAbssoluteVal=True)
    
    HeatList = []
    row_labels = []
    column_labels =[]

    if not Order:

        ps_copy = {i:'_'.join(reversed(v.split('_'))) for i, v in  PathSuffixDic.items()}

        
    else:

    
        ps_copy = {}
        for k,v in PathSuffixDic.items():


            if Order.split('-')[0] in v:
                ps_copy[k]=Order.split('-')[0]+v
            else:
                ps_copy[k]=Order.split('-')[1]+v

    
    
    
    for path, ky in sorted(ps_copy.items(),key=itemgetter(1)):
        
        
        ReacDic = GetPathFlux(res, {path:PathSuffixDic[path]}, XLSPath = None, X=None,Y=None, ShowLegend=False, ShowFig = False)
        
        if not MeanPath:
            for reac in ReacDic.keys():
                
                if ReacAbssoluteVal:
                     HeatList.append([abs(i) for i in ReacDic[reac]])
                else:
                     HeatList.append(ReacDic[reac])               
                row_labels.append(reac)
        else:
        
            scaleval=[np.absolute(i).mean() for i in zip(*ReacDic.values())]
            HeatList.append(scaleval)
            row_labels.append(path.split('_KEY_')[0]+PathSuffixDic[path])
            
    if len(ReacDic)==0:
        print("No reactions found")
        return None
    
    for n, col in enumerate(ReacDic[ReacDic.keys()[0]]):
        column_labels.append(n)
    row_labels.reverse()
    ax = sns.heatmap(HeatList, linewidths=linewidths)
    ax.set_yticklabels(row_labels, minor=False)
    ax.set_xticklabels(column_labels, minor=False)
    fig = ax.get_figure()
    plt.yticks(rotation=0)
    fig.show()
    

    


def PlotReacFlux(res, ReacListOrSuffix = None, XLSPath = None, X=None,Y=None, ShowLegend=False, ShowFig = False):

    ReacList = []
    
    for reac in res.to_dict().keys():
        if isinstance(ReacListOrSuffix,list):
            if reac in ReacListOrSuffix:
                ReacList.append(reac)
        if isinstance(ReacListOrSuffix,str):
            if ReacListOrSuffix in reac:
                ReacList.append(reac)
                
    Plot_Res(res,ReacList, XLSPath = XLSPath, X=X,Y=Y, ShowLegend=ShowLegend, ShowFig = ShowFig)

    return ReacList
        
    

        






def ROOM(model, refflux, reactions=None, delta=0, tol=0, IncZeroes=False,
         AsMtx=False, f=None, reset_state=True):
    """ refflux = {reac_name:flux} """
    if reset_state:
        state = model.GetState()
    model.ZeroObjective()
    if not reactions:
        reactions = model.Reactions()
    else:
        reactions = model.GetReactionNames(reactions)
    for reac in reactions:
        lb,ub = model.GetConstraint(reac)
        if reac in refflux:
            val = refflux[reac]
        else:
            val = 0.0
        hival = val + delta*abs(val) + tol
        loval = val - delta*abs(val) - tol
        model.AddMetabolite(reac+"_ROOMVarhi")
        model.SetMetBounds(reac+"_ROOMVarhi", -model.bounds*2, hival)
        model.AddMetabolite(reac+"_ROOMVarlo")
        model.SetMetBounds(reac+"_ROOMVarlo", loval, model.bounds*2)
        model.GetReaction(reac).add_metabolites({model.GetMetabolite(
            reac+"_ROOMVarhi"):1, model.GetMetabolite(reac+"_ROOMVarlo"):1})
        model.AddReaction(reac+"_ROOMVar",{model.GetMetabolite(
            reac+"_ROOMVarhi"):-(ub-hival), model.GetMetabolite(
            reac+"_ROOMVarlo"):-(lb-loval)}, bounds=(0,1))
        model.GetReaction(reac+"_ROOMVar").variable_kind = 'integer'
        model.SetObjective({reac+"_ROOMVar":1})
    model.SetObjDirec("Min")
    model.Solve(False)
    model.CleanUpTempVar("_ROOMVar")
    rv = model.GetSol(IncZeroes=IncZeroes,AsMtx=AsMtx,f=f)
    if reset_state:
        model.SetState(state, IncSol=False)
    return rv

def CleanUpROOM(model):
    model.CleanUpTempVar("_ROOMVar")


from cobra.manipulation import modify

def GeometricSol(model, IncZeroes=True, AsMtx=False, tol=1e-6, Print=False, cobra=False):
    state = model.GetState()
    reacs = model.reactions
    model.SetObjAsConstraint(name='FluxDiff'+str(0)+'_mindifffromflux')
    ### iteraion 1
    model.SetObjDirec("Min")
    it = 1
    model.ZeroObjective()
    modify.convert_to_irreversible(model)
    model.SetObjective(model.Reactions())
    model.Solve(False)
    model.SetObjAsConstraint(name='FluxDiff'+str(it)+'_mindifffromflux')
    fva = model.FVA(reaclist=reacs,tol=tol,PrintStatus=False, cobra=cobra)
    meanflux = fva.GetReacsMeanFlux(fva.keys())
    flux = dict(meanflux)
    flux_id = {}
    for reac in flux: 
    	flux_id[reac.id] = flux[reac]
    delta = fva.MaxDiff()
    variablereacs = fva.Variable(tol)
    variablereacs_id = [reac.id for reac in variablereacs]
    if Print:
        print("Iteration "+str(it)+", "+str(len(variablereacs))+" varible reactions, Max difference = "+str(delta))
    it += 1
    ### iteration n
    print(variablereacs)
    while (delta > tol):
        model.MinDiffFromFlux(flux_id, variablereacs_id, it, cleanup=False)
        model.SetObjAsConstraint(name='FluxDiff'+str(it)+'_mindifffromflux')
        fva = model.FVA(reaclist=variablereacs.keys(), tol=0,
                            PrintStatus=False, cobra=cobra)
        meanflux = fva.GetReacsMeanFlux(fva.keys())
        flux.update(meanflux)
        variablereacs = fva.Variable(tol)
        delta = fva.MaxDiff()
        if Print:
            print("Iteration "+str(it)+", "+str(len(variablereacs))+" varible reactions, Max difference = "+str(delta))
        it += 1
    ### cleanup
    model.CleanUpTempVar("_mindifffromflux")
    model.MergeRev()
    model.SetState(state)
    rv = model.GetSol(IncZeroes=True,AsMtx=False,sol=flux)
    return rv



import multiprocessing
from ..classes.matrix import matrix
from RWFMSolveMinFlux import RWFMSolveMinFlux #had to put function in diff .py because windows glitch
import time #added time for speed record

def RandomMinFlux(model,it=1,reacs=None,exc=[],processes=None):
    state = model.GetState()
    if reacs == None:
        reacs = model.Reactions()
    else:
        reacs = model.GetReactionNames(reacs)
    for reac in exc:
        reac = model.GetReactionName(exc)
        if reac in reacs:
            reacs.remove(reac)
    mtx = matrix()
    pool = multiprocessing.Pool(processes=processes)
    results = [pool.apply_async(RWFMSolveMinFlux, args=(model,)) for x in range(it)]
    pool.close()
    pool.join()
    sols = [x.get() for x in results]
    for sol in sols:
        mtx = mtx.UpdateFromDic(sol)
    model.SetState(state)
    end_time= time.time()- start_time
    
    print("Processing {} took {} seconds".format((model),end_time))
    return mtx




def DeadMetabolites(model, fva=None, reaclist=None, cobra=False):
    """ metabolites not involved in any allowed reactions """
    if fva == None:
        fva = model.FVA(reaclist=reaclist, cobra=cobra)
    allowed_mets =																																																					 []
    for r in fva.Allowed():
        for met in model.InvolvedWith(r):
            if model.GetMetaboliteName(met) not in allowed_mets:
                allowed_mets.append(model.GetMetaboliteName(met))
    return list(set(model.Metabolites()).difference(allowed_mets))

def NoDeadEndModel(model):
    tempmodel = model.Copy()
    pmets = tempmodel.PeripheralMetabolites()
    while len(pmets) > 0:
        tempmodel.DelMetabolites(pmets)
        pmets = tempmodel.PeripheralMetabolites()
    return tempmodel

def DeadEndMetabolites(model, nodeadendmodel=None):
    """ iteratively identify peripheral metabolites """
    if nodeadendmodel == None:
        nodeadendmodel = model.NoDeadEndModel()
    return list(set(model.Metabolites()).difference(nodeadendmodel.Metabolites()))

def PeripheralMetabolites(model, rc="all"):
    """ pre: rv ="all"|"Produced"|"Consumed"|"Orphan"
       post: returns dic or list of deadend metabolites """
    produced = []
    consumed = []
    orphan = []
    for met in model.Metabolites():
        prod = model.ProducedBy(met).keys()
        cons = model.ConsumedBy(met).keys()
        iw = list(set(prod).union(cons))
        if len(iw) == 1:
            orphan.append(met)
        if len(cons) == 0:
            produced.append(met)
        if len(prod) == 0:
            consumed.append(met)
    peri = list(set(produced).union(consumed).union(orphan))
    rvdic = {"all":peri,"Produced":produced,"Consumed":consumed,"Orphan":orphan}
    return rvdic[rc]

def ChokepointReactions(model):
    rv = []
    nodeadendmodel = model.NoDeadEndModel()
    for met in nodeadendmodel.Metabolites():
        prod = nodeadendmodel.ProducedBy(met)
        cons = nodeadendmodel.ConsumedBy(met)
        if len(prod) == 1:
            if prod.keys()[0] not in rv:
                rv.append(prod.keys()[0])
        if len(cons) == 1:
            if cons.keys()[0] not in rv:
                rv.append(cons.keys()[0])
    return model.GetReactionNames(rv)
    
def GetNeighbours(model, name, exclude=[]):
    rv = []
    nd = model.GetNeighboursAsDic(name,exclude)
    for neighbours in nd.values():
        for neighbour in neighbours:
            if neighbour not in rv:
                rv.append(neighbour)
    return rv

def GetNeighboursAsDic(model, name, exclude=[]):
    """  pre: name and exclude are reactions or metabolites names
        post: {iw:[neighbours]} """
    rv = {}
    iws = model.InvolvedWith(name,AsName=True)
    iws = list(set(iws).difference(exclude))
    for iw in iws:
        neighbours = model.InvolvedWith(iw, AsName=True).keys()
        neighbours = list(set(neighbours).difference([name]))
        neighbours = list(set(neighbours).difference(exclude))
        rv[iw] = neighbours
    return rv

def Degree(model, name, bipartite=True):
    if bipartite:
        n = model.InvolvedWith(name).keys()
    else:
        n = model.GetNeighbours(name)
    return len(n)

def DegreeDist(model, node_type="metabolites", bipartite=True):
    """ node_type = "metabolites" | "reactions" """
    rv = {}
    if 'met' in node_type:
        names = model.Metabolites()
    elif 'reac' in node_type:
        names = model.Reactions()
    for n in names:
        deg = model.Degree(n,bipartite=bipartite)
        if deg in rv.keys():
            rv[deg] = rv[deg] + 1
        else:
            rv[deg] = 1
    return rv

def MetabolitesDegree(model, mets=None, bipartite=True):
    rv = {}
    if not mets:
        mets = model.metabolites
    mets = model.GetMetabolites(mets)
    for met in mets:
        rv[met.id] = model.Degree(met, bipartite=bipartite)
    return rv

def ReactionsDegree(model, reacs=None, bipartite=True):
    rv = {}
    if not reacs:
        reacs = model.reactions
    reacs = model.GetReactions(reacs)
    for reac in reacs:
        rv[reac.id] = model.Degree(reac, bipartite=bipartite)
    return rv
